import {
  ChangeDetectionStrategy,
  Component,
  EventEmitter,
  Input,
  OnChanges,
  OnInit,
  Output,
  SimpleChanges,
  TemplateRef,
} from '@angular/core';
import { FormControl } from '@angular/forms';
import { debounceTime, distinctUntilChanged } from 'rxjs/operators';

import { DataTableColumn, DataTableColumnFormatType, FetchRecordsEvent } from './data-table.model';

@Component({
  selector: 'x-data-table',
  templateUrl: './data-table.component.html',
  styleUrls: ['./data-table.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush,
})
export class DataTableComponent implements OnInit, OnChanges {
  @Input() columns!: DataTableColumn[];
  @Input() records!: any[];
  @Input() displaySizeOptions!: any;
  @Input() totalRecordsCount = 0;
  @Input() pageSize = 10;
  @Input() label!: string;
  @Input() loading = false;
  @Input() selectable = false;
  @Input() selectedRecordsSet = new Set<any>();
  /**
   *
   * @description TemplateRef passed by name from parent component
   *
   * @param record Required \<ng-template\> attribute: let-record=record
   * @param column Optional \<ng-template\> attribute: let-column=column. Useful in case of multiple custom columns with 'templateRef' format type.
   *
   * @example
   * ```html
   * <ng-template #templateRefName
   *     let-column="column"
   *     let-record="record">
   *   <ng-container *ngIf="column.field === 'myColumnFieldName'">
   *     {{record.fleet.name}}
   *   </ng-container>
   * </ng-template>
   * ```
   */
  @Input() customCellTemplateRef?: TemplateRef<void>;
  @Output() fetchRecords: EventEmitter<FetchRecordsEvent> = new EventEmitter<FetchRecordsEvent>();
  @Output() selectRecord = new EventEmitter();

  readonly dataTableColumnFormatType = DataTableColumnFormatType;

  readonly PAGE_SIZE_OPTIONS = this.displaySizeOptions || [5, 10, 20, 50, 100];

  currentPageIndex = 1;
  currentPageStart = 0;
  currentPageEnd = 0;
  searchControl: FormControl = new FormControl('');
  sortedColumn: DataTableColumn | undefined;

  constructor() {}

  ngOnInit(): void {
    this._fetchRecords();
    this._subscribeToSearchControlChanges();
  }

  ngOnChanges(changes: SimpleChanges): void {
    if (
      changes.records.currentValue !== 'undefined' ||
      changes.totalRecordsCount.currentValue !== 'undefined'
    ) {
      this.loading = false;
      this._setPaginationStartAndEnd();
    }
  }

  reset() {
    this._resetPagination();
    if (this.searchControl.value) {
      this.clearSearch();
    } else {
      this._fetchRecords();
    }
  }

  clearSearch() {
    this.searchControl.setValue('');
    this.searchControl.updateValueAndValidity();
  }

  /* Handle Selection Checkboxes */
  updateCheckedSet(id: number, checked: boolean): void {
    if (checked) {
      this.selectedRecordsSet.add(id);
    } else {
      this.selectedRecordsSet.delete(id);
    }
  }

  onAllChecked(checked: boolean): void {
    this.records.forEach((rec) => this.updateCheckedSet(rec.id, checked));
    this.selectRecord.emit(this.selectedRecordsSet);
  }

  onItemChecked(id: number, checked: boolean): void {
    this.updateCheckedSet(id, checked);
    this.selectRecord.emit(this.selectedRecordsSet);
  }

  /* Handle Pagination */
  onPageSizeChanged(size: number) {
    this.pageSize = size;
    this.currentPageIndex = 1;
    this._fetchRecords();
  }

  onPageIndexChanged(pageIndex: number) {
    this.currentPageIndex = pageIndex;
    this._fetchRecords();
  }

  private _subscribeToSearchControlChanges() {
    this.searchControl.valueChanges
      .pipe(debounceTime(300), distinctUntilChanged())
      .subscribe(() => {
        this._resetPagination();
        this._fetchRecords();
      });
  }
  private _resetPagination() {
    this.currentPageIndex = 1;
    this.currentPageStart = 0;
    this.currentPageEnd = 0;
  }

  private _fetchRecords() {
    this.loading = true;
    this.fetchRecords.emit({
      searchKey: this.searchControl.value,
      /* @TODO Implement sorting or remove this */
      //sortOrder: this._getSortOrder(),
      //sortField: this.sortedColumn?.field,
      paginationInfo: { pageNumber: this.currentPageIndex, pageSize: this.pageSize },
    });
  }

  private _setPaginationStartAndEnd() {
    const start = (this.currentPageIndex - 1) * this.pageSize;
    this.currentPageStart = this.totalRecordsCount ? start + 1 : start;
    this.currentPageEnd = start + this.pageSize;
    this.currentPageEnd =
      this.currentPageEnd > this.totalRecordsCount ? this.totalRecordsCount : this.currentPageEnd;
  }

  /* @TODO Implement sorting or remove this */
  // onSortingChanged(sortedColumn: DataTableColumn) {
  //   this._resetPagination();
  //   this.sortedColumn = sortedColumn;
  //   this._fetchRecords();
  // }

  // private _getSortOrder(): string | undefined {
  //   if (this.sortedColumn?.sortOrder) {
  //     return this.sortedColumn.sortOrder === 'ascend' ? 'ASC' : 'DESC';
  //   }
  //   return undefined;
  // }
}
